/*
 * Copyright by Intland Software
 *
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Intland Software. ("Confidential Information"). You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Intland.
 */

package com.intland.codebeamer.importer.bugzilla;

import java.io.IOException;
import java.io.Reader;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;

import com.ibatis.common.resources.Resources;
import com.ibatis.sqlmap.client.SqlMapClient;
import com.ibatis.sqlmap.client.SqlMapClientBuilder;
import com.intland.codebeamer.importer.bugzilla.dao.BugzillaDaoImpl;
import com.intland.codebeamer.importer.bugzilla.dao.QueryIdProvider;
import com.intland.codebeamer.manager.AccessRightsException;
import com.intland.codebeamer.persistence.dto.AssociationDto;
import com.intland.codebeamer.persistence.dto.BinaryStreamDto;
import com.intland.codebeamer.persistence.dto.ProjectDto;
import com.intland.codebeamer.persistence.dto.TrackerChoiceOptionDto;
import com.intland.codebeamer.persistence.dto.TrackerDto;
import com.intland.codebeamer.persistence.dto.TrackerItemAttachmentDto;
import com.intland.codebeamer.persistence.dto.TrackerItemCommentDto;
import com.intland.codebeamer.persistence.dto.TrackerItemDto;
import com.intland.codebeamer.persistence.dto.TrackerItemHistoryEntryDto;
import com.intland.codebeamer.persistence.dto.TrackerLayoutLabelDto;
import com.intland.codebeamer.persistence.dto.UserDto;
import com.intland.codebeamer.persistence.util.BinaryStreamDtoHelper;
import com.intland.codebeamer.remoting.GroupType;
import com.intland.codebeamer.remoting.RemoteApi;
import com.intland.codebeamer.remoting.RemoteApiFactory;

/**
 * @author <a href="mailto:robert.enyedi@intland.com">Robert Enyedi</a>
 * @version $Id$
 */
public class Importer {

	public static final String BUGZILLA_OLD_VERSION_PARAMETER = "version=old";

	protected static Logger logger = Logger.getLogger(Importer.class);

	private SqlMapClient sqlMap;
	protected RemoteApi api;
	protected String token;

	protected Map usersMap = new HashMap();

	protected Map severitiesMap = new HashMap();
	protected Map opsysMap = new HashMap();
	protected Map platformsMap = new HashMap();
	protected Map categoriesMap = new HashMap();
	protected Map resolutionsMap = new HashMap();
	protected Map milestonesMap = new HashMap();
	protected Map versionsMap = new HashMap();
	protected Map statusesMap = new HashMap();

	protected boolean oldBugzillaVersion = false;

	/**
	 * This map holds the associations between a Bugzilla ID and its corresponding CB tracker item Id.
	 */
	public Map/*<Integer, Integer>*/ trackersByBugzillaIdMap = new HashMap();

	protected Map trackerFieldMappings = new HashMap();

	private QueryIdProvider queryIdProvider;

	static {
		// Init log4j
		try {
			PropertyConfigurator.configure(Resources.getResourceURL("com.intland.codebeamer/importer/bugzilla/log4j.properties"));
		} catch (IOException ex) {
			ex.printStackTrace();
			System.err.println("Error initializing log4j");
		}
	}

	public SqlMapClient getSqlMapInstance() {
		return sqlMap;
	}

	protected void initIBatis() {
		// Init iBATIS
		try {
			String resource = "com.intland.codebeamer/importer/bugzilla/sqlmap-config.xml";
			Reader reader = Resources.getResourceAsReader(resource);
			sqlMap = SqlMapClientBuilder.buildSqlMapClient(reader);
		} catch (Exception ex) {
			ex.printStackTrace();
			throw new RuntimeException(
					"Error initializing iBATIS. Cause: " + ex);
		}
	}

	public static void main(String[] args) throws Exception {

		if(args.length < 3 || args.length > 5) {
			displayUsage();
		} else {
			String serviceUrl = args[0];
			String login = args[1];
			String password = args[2];
			boolean oldVersion = false;

			String product = null;

			if(args.length > 4) {
				if(args[3].equalsIgnoreCase(BUGZILLA_OLD_VERSION_PARAMETER)) {
					oldVersion = true;

					logger.info("Importer configured for old Bugzilla version.");

					if(args.length ==5 ) {
						product = args[4];
					}
				} else {
					product = args[3];
				}
			}

			new Importer().start(serviceUrl, login, password, oldVersion, product);
		}
	}

	protected static void displayUsage() {
		System.err.println("Usage: <program> service-URL login password [" + BUGZILLA_OLD_VERSION_PARAMETER + "] [productName]");
		System.err.println("\nSpecify the \"" + BUGZILLA_OLD_VERSION_PARAMETER + "\" parameter if your Bugzilla version is older than 2.16.");
		System.err.println("The optional \"productName\" argument is used for importing a single product at a time.");
		System.err.println("\nSample program arguments: http://localhost:8080/cb/remote-api bond 007");
	}

	public void start(String serviceUrl, String login, String password, boolean oldVersion, String product) throws MalformedURLException {

		this.oldBugzillaVersion = oldVersion;

		initIBatis();
		queryIdProvider = new QueryIdProvider(this);

		api = RemoteApiFactory.getInstance().connect(serviceUrl);
		token = api.login(login, password);

		// TODO Check server version to be 4.2+

		try {
			importUsers();
			importProjects(product);
		} catch (Exception ex) {
			logger.error(ex,ex);
		}
	}

	private void importTrackerLabels(TrackerDto bugTracker) throws Exception {

			bugTracker.setUsingWorkflow(false);
			api.updateTracker(token, bugTracker);

			Importer.logger.debug("Removing default label option values");
			api.deleteTrackerChoiceOptions(token, bugTracker.getId());

			importTrackerCategories(bugTracker);
			importTrackerOpSys(bugTracker);
			importTrackerPlatforms(bugTracker);
			importTrackerResolutions(bugTracker);
			importTrackerSeverities(bugTracker);

			importTrackerStatuses(bugTracker);

			importTrackerMilestones(bugTracker);
			importTrackerVersions(bugTracker);
			importTrackerKeywords(bugTracker);

			adjustCustomFieldNames(bugTracker);

			registerTrackerFieldMappings(bugTracker);
	}

	protected void adjustCustomFieldNames(TrackerDto bugTracker) throws AccessRightsException {

		TrackerLayoutLabelDto bugzillaIdLabel = Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.getCustomFieldId(0));
		bugzillaIdLabel.setLabel("Bugzilla ID");
		Utils.updateLatestTrackerLayout(this);
	}

	/** Defines the correlation between the Bugzilla field names and CB field types */
	private void registerTrackerFieldMappings(TrackerDto bugTracker) {

		trackerFieldMappings.clear();

		trackerFieldMappings.put("version", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.VERSION_LABEL_ID));
		trackerFieldMappings.put("rep_platform", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.PLATFORM_LABEL_ID));
		trackerFieldMappings.put("op_sys", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.OP_SYS_LABEL_ID));
		trackerFieldMappings.put("bug_status", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.STATUS_LABEL_ID));
		trackerFieldMappings.put("resolution", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.RESOLUTION_LABEL_ID));
		trackerFieldMappings.put("bug_severity", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.SEVERITY_LABEL_ID));
		trackerFieldMappings.put("priority", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.PRIORITY_LABEL_ID));
		trackerFieldMappings.put("version", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.VERSION_LABEL_ID));
		trackerFieldMappings.put("component", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.CATEGORY_LABEL_ID));
		trackerFieldMappings.put("version", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.VERSION_LABEL_ID));
		trackerFieldMappings.put("assigned_to", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.ASSIGNED_TO_LABEL_ID));
		trackerFieldMappings.put("reporter", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.SUBMITTED_BY_LABEL_ID));
		trackerFieldMappings.put("version", Utils.getFieldByType(this, bugTracker, TrackerLayoutLabelDto.VERSION_LABEL_ID));
//		trackerFieldMappings.put("qa_contact", Utils.getFieldByType(this, bugTracker, APPROVER?));
	}

	protected TrackerDto findProjectBugTracker(ProjectDto project) {
		TrackerDto[] trackers = api.findTrackersByProject(token, project.getId());
		TrackerDto bugTracker = null;

		if(trackers == null || trackers.length == 0) {
			logger.fatal("Project contains no trackers.");
			return null;
		}

		// Locate the bug tracker
		for(int i=0; i<trackers.length; i++) {
			// TODO This is heuristic
			if(trackers[i].getName().toLowerCase().indexOf("bug") != -1) {
				bugTracker = trackers[i];

				logger.debug("Identified bug tracker: " + bugTracker);
				break;
			}
		}

		if(bugTracker == null) {
			bugTracker = trackers[1];
			logger.debug("Failed to locate a bug tracker. Using first tracker instead: " + bugTracker);
		}

		return bugTracker;
	}

	private void importTrackerOpSys(TrackerDto bugTracker) throws Exception {
		List operatingSystems = BugzillaDaoImpl.getInstance().findAllOpSys(this, bugTracker.getProject());

		logger.debug("Operating systems: " + operatingSystems);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, operatingSystems, TrackerLayoutLabelDto.OP_SYS_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, opsysMap, TrackerLayoutLabelDto.OP_SYS_LABEL_ID);
	}

	private void importTrackerPlatforms(TrackerDto bugTracker) throws Exception {
		List platforms = BugzillaDaoImpl.getInstance().findAllPlatforms(this, bugTracker.getProject());

		logger.debug("Platforms: " + platforms);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, platforms, TrackerLayoutLabelDto.PLATFORM_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, platformsMap, TrackerLayoutLabelDto.PLATFORM_LABEL_ID);
	}

	private void importTrackerCategories(TrackerDto bugTracker) throws Exception {
		List categories = BugzillaDaoImpl.getInstance().findAllCategories(this, bugTracker.getProject());

		logger.debug("Categories: " + categories);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, categories, TrackerLayoutLabelDto.CATEGORY_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, categoriesMap, TrackerLayoutLabelDto.CATEGORY_LABEL_ID);
	}

	private void importTrackerResolutions(TrackerDto bugTracker) throws Exception {
		List resolutions = BugzillaDaoImpl.getInstance().findAllResolutions(this, bugTracker.getProject());

		logger.debug("Resolutions: " + resolutions);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, resolutions, TrackerLayoutLabelDto.RESOLUTION_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, resolutionsMap, TrackerLayoutLabelDto.RESOLUTION_LABEL_ID);
	}

	private void importTrackerSeverities(TrackerDto bugTracker) throws Exception {
		List severities = BugzillaDaoImpl.getInstance().findAllSeverities(this, bugTracker.getProject());

		logger.debug("Severities: " + severities);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, severities, TrackerLayoutLabelDto.SEVERITY_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, severitiesMap, TrackerLayoutLabelDto.SEVERITY_LABEL_ID);
	}

	private void importTrackerStatuses(TrackerDto bugTracker) throws Exception {
		List statuses = BugzillaDaoImpl.getInstance().findAllStatuses(this, bugTracker.getProject());

		logger.debug("Statuses: " + statuses);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, statuses, TrackerLayoutLabelDto.STATUS_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, statusesMap, TrackerLayoutLabelDto.STATUS_LABEL_ID);
	}

	private void importTrackerMilestones(TrackerDto bugTracker) throws Exception {
		List milestones = BugzillaDaoImpl.getInstance().findAllMilestones(this, bugTracker.getProject());

		logger.debug("Milestones: " + milestones);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, milestones, TrackerLayoutLabelDto.MILESTONES_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, milestonesMap, TrackerLayoutLabelDto.MILESTONES_LABEL_ID);
	}

	private void importTrackerVersions(TrackerDto bugTracker) throws Exception {
		List versions = BugzillaDaoImpl.getInstance().findAllVersions(this, bugTracker.getProject());

		logger.debug("Versions: " + versions);

		Utils.importTrackerLabelChoiceOptions(this, bugTracker, versions, TrackerLayoutLabelDto.VERSION_LABEL_ID);
		Utils.buildValuesMap(this, bugTracker, versionsMap, TrackerLayoutLabelDto.VERSION_LABEL_ID);
	}

	private void importTrackerKeywords(TrackerDto bugTracker) throws Exception {
		List keywords = BugzillaDaoImpl.getInstance().findAllKeywords(this, bugTracker.getProject());

		logger.debug("Keywords: " + keywords);

		// TODO What about the keyword descriptions?
		Utils.importTrackerLabelChoiceOptions(this, bugTracker, "Keywords", keywords, TrackerLayoutLabelDto.getChoiceFieldId(0));
	}

	public void importUsers() throws Exception {

		List list = BugzillaDaoImpl.getInstance().findAllUsers(this);

		logger.info("Importing " + list.size() + " users...");

		usersMap.clear();

		for(Iterator it = list.iterator(); it.hasNext(); ) {
			UserDto crtUser = (UserDto) it.next();

			Utils.splitName(crtUser);

			logger.debug("Creating user " + crtUser);

			crtUser = api.createUser(token, crtUser);

			usersMap.put(crtUser.getName().toLowerCase(), crtUser);
		}

		logger.info("Users successfully imported.");
	}

	public void importProjects(String product) throws Exception {
		List list = BugzillaDaoImpl.getInstance().findAllProjects(this);

		trackersByBugzillaIdMap.clear();

		if(product != null) {
			logger.info("Single product processing mode for: " + product);
		}

		for(int i=0; i<list.size(); i++) {
			ProjectDto crtProject = (ProjectDto) list.get(i);

			if(product == null || (product != null && crtProject.getName().equals(product)))
			{
				ProjectDto tmpProject = api.findProjectByName(token, crtProject.getName());

				logger.info("Processing project " + "[" + (i+1) + "/" + list.size() + "]: " + crtProject.getName());
				logger.info("====================================================");

				if(tmpProject != null) {
					logger.info("Using existing project: " + tmpProject);
					crtProject = tmpProject;
				} else {
					logger.info("Creating project " + crtProject.getName());
					crtProject = api.createProject(token, crtProject);
				}

				TrackerDto bugTracker = findProjectBugTracker(crtProject);

				if(bugTracker != null) {

					logger.info("Using bug tracker: " + bugTracker);

					importTrackerLabels(bugTracker);
					importTrackerData(bugTracker);
				} else {
					logger.fatal("Could not identify bug tracker!");
				}
			}
		}

		if(product != null) {
			logger.warn("The setting of item associations and duplicates for inter-product references does not work in case of single product imports.");
		}

		importTrackerItemAssociations();
		markDuplicateBugs();

		logger.info("Import completed successfully.");
	}

	protected void markDuplicateBugs() throws Exception {

		logger.info("Marking duplicate bugs...");

		List/*<Map>*/ duplicates = BugzillaDaoImpl.getInstance().findAllDuplicates(this);

		for(Iterator it = duplicates.iterator(); it.hasNext(); ) {
			Map crtDuplicate = (Map) it.next();

			Integer duplicateInteger = (Integer) crtDuplicate.get("dupe");
			Integer duplicateOfInteger = (Integer) crtDuplicate.get("dupe_of");

			if(duplicateInteger != null && duplicateOfInteger != null) {

				Integer duplicateItemId = (Integer) trackersByBugzillaIdMap.get(duplicateInteger.toString());
				Integer duplicateOfItemId = (Integer) trackersByBugzillaIdMap.get(duplicateOfInteger.toString());

				if(duplicateItemId != null && duplicateOfItemId != null) {
					logger.debug("Marking item " + duplicateItemId + " as duplicate of " + duplicateOfItemId);

					// Creating an association between the items
					AssociationDto association = new AssociationDto();
					association.setFromId(duplicateItemId);
					association.setFromTypeId(new Integer(GroupType.TRACKER_ITEM));
					association.setToId(duplicateOfItemId);
					association.setToTypeId(new Integer(GroupType.TRACKER_ITEM));
					association.setSubmitter(api.getSessionUser(token));

					association = api.createAssociation(token, association);

					// Attempting to mark item as duplicate
					TrackerItemDto duplicateItem = api.findTrackerItemById(token, duplicateItemId);

					if(duplicateItem != null) {
						TrackerDto bugTracker = findProjectBugTracker(duplicateItem.getTracker().getProject());
						TrackerChoiceOptionDto duplicateResolutionOption = Utils.findResolutionForDuplicate(this, bugTracker);

						if(duplicateResolutionOption != null) {
							List resolutionsList = new ArrayList();
							resolutionsList.add(duplicateResolutionOption);
							duplicateItem.setResolutions(resolutionsList);

							api.updateTrackerItem(token, duplicateItem);

							logger.debug("Tracker item " + duplicateItemId + " successfully marked as duplicate of " + duplicateOfItemId);
						}
					} else {
						logger.error("Cannot find tracker item with ID " + duplicateItemId);
					}
				} else {
					logger.error("Could not find tracker items for Bugzilla IDs: dupe = " + duplicateInteger + ", dupe_of = " + duplicateOfInteger);
				}
			} else {
				logger.error("Invalid bug duplication record detected: dupe = " + duplicateInteger + ", dupe_of = " + duplicateOfInteger);
			}
		}
	}

	private void importTrackerData(TrackerDto bugTracker) throws Exception {
		List/*<TrackerItemDto>*/ trackerItems = BugzillaDaoImpl.getInstance().findAllBugsData(this, bugTracker);

		logger.info("Importing " + trackerItems.size() + " tracker items...");

		for(Iterator it = trackerItems.iterator(); it.hasNext(); ) {
			TrackerItemDto crtTrackerItem = (TrackerItemDto) it.next();

			Date modifiedAtDate = crtTrackerItem.getModifiedAt();
			UserDto modifiedBy = crtTrackerItem.getModifier();

			logger.debug("Creating tracker item " + crtTrackerItem);

			crtTrackerItem = api.createTrackerItem(token, crtTrackerItem);

			importTrackerItemAttachments(crtTrackerItem);
			importTrackerItemHistory(crtTrackerItem);
			importTrackerItemComments(crtTrackerItem);

			// Restore the original modified_at date and modifier
			crtTrackerItem.setModifiedAt(modifiedAtDate);
			crtTrackerItem.setModifier(modifiedBy);
			api.updateTrackerItemAsModified(token, crtTrackerItem);

			trackersByBugzillaIdMap.put(getBugzillaId(crtTrackerItem), crtTrackerItem.getId());
		}

		logger.info("Tracker items successfully imported.");
		logger.info("====================================================");
	}

	private void importTrackerItemAssociations() throws Exception {

		List/*<Map>*/ dependencies = BugzillaDaoImpl.getInstance().findAllDependencies(this);

		if(dependencies.size() > 0) {
			logger.info("Creating tracker item associations...");
		} else {
			logger.info("No bug dependencies found.");
		}

		for(Iterator it = dependencies.iterator(); it.hasNext(); ) {
			Map crtDependency = (Map) it.next();

			Integer blockedInteger = (Integer) crtDependency.get("blocked");
			Integer dependsOnInteger = (Integer) crtDependency.get("dependson");

			if(blockedInteger != null && dependsOnInteger != null) {

				Integer blockedItemId = (Integer) trackersByBugzillaIdMap.get(blockedInteger.toString());
				Integer dependsOnItemId = (Integer) trackersByBugzillaIdMap.get(dependsOnInteger.toString());

				if(blockedItemId != null && dependsOnItemId != null) {
					logger.debug("Creating association between " + blockedItemId + " and " + dependsOnItemId);

					AssociationDto association = new AssociationDto();
					association.setFromId(blockedItemId);
					association.setFromTypeId(new Integer(GroupType.TRACKER_ITEM));
					association.setToId(dependsOnItemId);
					association.setToTypeId(new Integer(GroupType.TRACKER_ITEM));
					association.setSubmitter(api.getSessionUser(token));

					association = api.createAssociation(token, association);
				} else {
					logger.error("Could not find tracker items for Bugzilla IDs: blocked = " + blockedInteger + ", depends on = " + dependsOnInteger);
				}
			} else {
				logger.error("Invalid bug association detected: blocked = " + blockedInteger + ", depends on = " + dependsOnInteger);
			}
		}
	}

	private void importTrackerItemHistory(TrackerItemDto crtTrackerItem) throws Exception {

		List/*<Map>*/ historyEntries = BugzillaDaoImpl.getInstance().findAllHistoryForBug(this, getBugzillaId(crtTrackerItem));

		for(Iterator it = historyEntries.iterator(); it.hasNext(); ) {
			Map crtHistoryEntry = (Map) it.next();

			logger.debug("Creating history entry");

			TrackerItemHistoryEntryDto historyEntry = new TrackerItemHistoryEntryDto();
			historyEntry.setTrackerItem(crtTrackerItem);

			String submitterString = (String) crtHistoryEntry.get("who");

			if(submitterString != null) {
				UserDto submittedBy = (UserDto) usersMap.get(submitterString.toLowerCase());

				if(submittedBy != null) {
					historyEntry.setSubmitter(submittedBy);
				} else {
					logger.warn("Could not find submitter " + submitterString);
				}
			}

			String fieldNameString = (String) crtHistoryEntry.get("fieldname");
			TrackerLayoutLabelDto labelDto = (TrackerLayoutLabelDto) trackerFieldMappings.get(fieldNameString);

			historyEntry.setSubmittedAt((Date) crtHistoryEntry.get("bug_when"));

			if(labelDto != null) {
				historyEntry.setLabel(labelDto);

				historyEntry.setNewValue((String) crtHistoryEntry.get("added"));
				historyEntry.setOldValue((String) crtHistoryEntry.get("removed"));

				api.createTrackerItemHistoryEntry(token, historyEntry);
			} else {
				logger.debug("Could not find tracker layout label for Bugzilla field name fro history: " + fieldNameString);
			}
		}
	}

	private void importTrackerItemComments(TrackerItemDto crtTrackerItem) throws Exception {
		List/*Map*/ comments = BugzillaDaoImpl.getInstance().findAllCommentsForBug(this, getBugzillaId(crtTrackerItem));

		for(Iterator it = comments.iterator(); it.hasNext(); ) {

			logger.debug("Creating comment");

			Map crtComment = (Map) it.next();

			TrackerItemCommentDto comment = new TrackerItemCommentDto();
			comment.setTrackerItem(crtTrackerItem);
			comment.setSubmittedAt((Date) crtComment.get("bug_when"));

			String submitterString = (String) crtComment.get("who");

			if(submitterString != null) {
				UserDto submittedBy = (UserDto) usersMap.get(submitterString.toLowerCase());

				if(submittedBy != null) {
					comment.setSubmitter(submittedBy);
				} else {
					logger.warn("Could not find submitter " + submitterString);
				}
			}

			comment.setDescription((String) crtComment.get("thetext"));

			api.createTrackerItemComment(token, comment);
		}
	}

	private void importTrackerItemAttachments(TrackerItemDto crtTrackerItem) throws Exception {
		// Add attachments
		List/*Map*/ attachments = BugzillaDaoImpl.getInstance().findAllAttachmentsForBug(this, getBugzillaId(crtTrackerItem));

		for(Iterator attachmentsIt = attachments.iterator(); attachmentsIt.hasNext(); ) {

			logger.debug("Creating attachment");

			Map crtAttachment = (Map) attachmentsIt.next();

			TrackerItemAttachmentDto attachment = new TrackerItemAttachmentDto();

			attachment.setTrackerItem(crtTrackerItem);
			attachment.setDescription((String) crtAttachment.get("description"));
			attachment.setMimeType((String) crtAttachment.get("mimetype"));
			attachment.setFilename((String) crtAttachment.get("filename"));
			attachment.setSubmittedAt((Date) crtAttachment.get("creation_ts"));

			BinaryStreamDto attachmentStream = BinaryStreamDtoHelper.createFromBytes((byte[]) crtAttachment.get("thedata"));

			String submitterString = (String) crtAttachment.get("submitter");

			if(submitterString != null) {
				UserDto submittedBy = (UserDto) usersMap.get(submitterString.toLowerCase());

				if(submittedBy != null) {
					attachment.setSubmitter(submittedBy);
				} else {
					logger.warn("Could not find submitter " + submitterString);
				}
			}

			attachment = api.createTrackerItemAttachment(token, attachment);
			if (attachmentStream != null) {
				api.saveTrackerItemAttachmentBlob(token, attachment.getId(), attachmentStream);
			}
		}
	}

	public static String getBugzillaId(TrackerItemDto crtTrackerItem) {
		return crtTrackerItem.getCustomField(0);
	}

	public RemoteApi getApi() {
		return api;
	}

	public String getToken() {
		return token;
	}

	public Map getCategoriesMap() {
		return categoriesMap;
	}

	public Map getMilestonesMap() {
		return milestonesMap;
	}

	public Map getOpsysMap() {
		return opsysMap;
	}

	public Map getPlatformsMap() {
		return platformsMap;
	}

	public Map getResolutionsMap() {
		return resolutionsMap;
	}

	public Map getSeveritiesMap() {
		return severitiesMap;
	}

	public Map getStatusesMap() {
		return statusesMap;
	}

	public Map getUsersMap() {
		return usersMap;
	}

	public Map getVersionsMap() {
		return versionsMap;
	}

	public boolean isOldBugzillaVersion() {
		return oldBugzillaVersion;
	}

	public QueryIdProvider getQueryIdProvider() {
		return queryIdProvider;
	}

}
